--[[ Roster

	This module handles the list of pets that will be displayed in the browser.
	The actual displaying is handled by the pet panel and pet filter menu.

	Instead of relying on the craziness of the default journal, Rematch 4.0 will
	work off its own list of pets.  The only time the journal is involved is when
	gathering a list of all pets to pick up petIDs or to grab source categories.

	settings.Filters = {
		Search = "",
		Collected = { Owned=boolean, Missing=boolean },
		Types = { booleanx10 },
		Strong = { booleanx10 },
		Tough = { booleanx10 },
		Sources = { booleanx10 },
		Rarity = { booleanx4 },
		Level = { 1(levels 1-7), 2(8-14), 3(15-24), 4(25), Without25s, MovesetNot25 all boolean},
		Other = { Favorite, NotFavorite, Leveling, NotLeveling, Tradable, NotTradable,
							Battle, NotBattle, Qty1, Qty2, Qty3, InTeam, NotInTeam },
	}
	settings.Sort = {
		Order = 1-7, (1=name, 2=level, 3=rarity, 4=type, 5=health, 6=power, 7=speed)
		Reverse = boolean,
		FavoritesFirst = boolean,
	}
	boolean are actually true/nil so next(table) will report whether it's empty
]]

local _,L = ...
local rematch = Rematch
local settings
rematch.Roster = {}
local roster = rematch.Roster

roster.updatingSelf = nil -- becomes true when roster is updating itself (to prevent reacting to own events)
roster.needsUpdated = true -- becomes true when we need to expand filters to grab all pets

-- place to backup and restore journal settings
roster.journalBackup = { search="", collected=nil, notCollected=nil, types={}, sources={} }

roster.pets = {} -- master list of all owned petIDs and missing speciesIDs
roster.list = {} -- list of filtered pets to display

roster.speciesSources = {} -- source indexes of each species, created if source filter ever used
roster.speciesAt25 = {} -- indexed by speciesID, the species that have at least one pet at 25
roster.searchMask = nil -- desensitized text to search for ([xX]51%-[nN][eE]etc)
roster.searchStatRanges = {} -- Level={1,24} Speed={300,305} Health={1400,nil} Power={nil,100}

-- sort optimization
roster.sortStats = {} -- indexed by petID/speciesID, the stat currently sorting by (level, rarity, type, etc)
roster.favorites = {} -- indexed by petID, true for pets that are currently flagged as favorite
roster.sortLowToHigh = nil -- defined before a sort begins, whether a sort lists in ascending or descending stat value

-- search/filter optimization and only contain data for the life of a search
roster.movesetsAt25 = {} -- indexed by a moveset "123,456,789" of a 25 pet (only contains data while filter enabled)
roster.cacheMoveset = {} -- indexed by speciesID, true/false when a moveset is 25 (only contains data while filter enabled)
roster.cacheStrong = {} -- indexed by speciesID, true/false when a species determined to be strong (only contains data while filter enabled)
roster.cacheSearchSpecies = {} -- indexed by speciesID, true/false when a species' abilities matches the searchMask
roster.cacheSearchAbilities = {} -- indexed by abilityID, true/false when an ability matches the searchMask

roster.uniquePets = 0 -- tallied when number of owned pets changes, the number of unique pets owned

-- settings.Filter table keys and their localized name
roster.filterGroups = { Collected=COLLECTED, Favorite=L["Favorites"], Types=L["Types"],
												Strong=L["Strong Vs"], Tough=L["Tough Vs"],	Sources=SOURCES, Rarity=RARITY,
												Breed=L["Breed"], Level=LEVEL, Other=OTHER, Similar=L["Similar"] }

-- radio group definitions: variables that are mutually exclusive with each other
roster.radioGroups = {
	Without25s=L["Level"], MovesetNot25=L["Level"], Leveling=L["Leveling"], NotLeveling=L["Leveling"],
	Tradable=L["Tradable"], NotTradable=L["Tradable"], Battle=L["Battle"], NotBattle=L["Battle"],
	Qty1=L["Quantity"], Qty2=L["Quantity"], Qty3=L["Quantity"], InTeam=L["Team"], NotInTeam=L["Team"] }

roster.searchStatMasks = { [PET_BATTLE_STAT_HEALTH:lower()]="Health", [PET_BATTLE_STAT_POWER:lower()]="Power",
	[PET_BATTLE_STAT_SPEED:lower()]="Speed", [LEVEL:lower()]="Level", ["health"]="Health", ["power"]="Power",
	["speed"]="Speed", ["level"]="Level", ["h"]="Health", ["p"]="Power", ["s"]="Speed", ["l"]="Level" }

rematch:InitModule(function()
	settings = RematchSettings
	-- make sure each filterGroup has a table in settings.Filters
	for var in pairs(roster.filterGroups) do
		if type(settings.Filters[var])~="table" then
			settings.Filters[var] = {}
		end
		-- if ResetFilters option enabled, clear all filters on login
		if settings.ResetFilters then
			wipe(settings.Filters[var])
		end
	end
	-- if a sort isn't defined (or ResetFilters+ResetSortWithFilters both enabled), create an empty one
	if settings.ResetSortWithFilters and settings.ResetFilters then
		roster:ClearSort()
	end
end)

--[[ stuff called from outside ]]

-- returns number of filtered pets, or number of all pets if all is true
function roster:GetNumPets(all)
	return all and #roster.pets or #roster.list
end

-- returns a petID at an index of the filtered list, or of all pets if all is true
function roster:GetPetByIndex(index,all)
	return all and roster.pets[index] or roster.list[index]
end

-- sets a filter: SetFilter("collected","owned",true)
function roster:SetFilter(group,variable,value)
	settings.Filters[group][variable] = value or nil
end

-- returns the state of a filter: GetFilter("sources",3)
function roster:GetFilter(group,variable)
	return settings.Filters[group][variable]
end

-- returns whether a filter group is clear: IsFilterClear("types")
function roster:IsFilterClear(group)
	return not next(settings.Filters[group])
end

-- returns true if a filter is being used (has something inside)
function roster:IsFilterUsed(group)
	return next(settings.Filters[group]) and true
end

-- returns true if a numerically-indexed filter is full
function roster:IsFilterFull(group,size)
	for i=1,size do
		if not settings.Filters[group][i] then
			return false
		end
	end
	return true
end

-- clears a filter group: ClearFilter("rarity")
function roster:ClearFilter(group)
	wipe(settings.Filters[group])
end

-- clears a filter radio group: ClearFilter("Other","Quantity")
function roster:ClearRadio(group,radioGroup)
	local filter = settings.Filters[group]
	if filter then
		for k,v in pairs(roster.radioGroups) do
			if v==radioGroup then
				filter[k] = nil
			end
		end
	end
end

-- clears all filters
function roster:ClearAllFilters()
	for k,v in pairs(settings.Filters) do
		if type(v)=="table" and roster.filterGroups[k] then
			wipe(v)
		else
			settings.Filters[k] = nil
		end
	end
	-- if ResetSortWithFilters enabled, also clear sort
	if settings.ResetSortWithFilters then
		roster:ClearSort()
	end
	-- clearing search via petpanel search box
	rematch.PetPanel.Top.SearchBox.Clear:Click()
	rematch:UpdateRoster()
	rematch:HideDialog()
end

function roster:GetSort(var)
	return settings.Sort[var]
end

function roster:SetSort(var,value)
	settings.Sort[var] = value or nil
end

function roster:ClearSort()
	settings.Sort = { Order=1, FavoritesFirst=true }
end

-- returns a comma-separated list of filters used for displaying on petpanel header; or nil if no filters used
-- pass groupsOnly as true when a list of filters groups only should be returned (ie favorite filters)
function roster:GetFiltersText(groupsOnly)
	local info = rematch.info
	wipe(info)
	if not groupsOnly then
		if roster.searchMask then
			tinsert(info,L["Search"])
		end
		-- if ResetSortWithFilters enabled, also show when a non-default Sort is active
		if settings.ResetSortWithFilters and (settings.Sort.Order~=1 or settings.Sort.Reverse or not settings.Sort.FavoritesFirst) then
			tinsert(info,L["Sort"])
		end
	end
	for filter,filterName in pairs(roster.filterGroups) do
		if roster:IsFilterUsed(filter) then
			if filter=="Other" then
				for otherFilter,otherFilterName in pairs(roster.radioGroups) do
					if roster:GetFilter(filter,otherFilter) then
						tinsert(info,otherFilterName)
					end
				end
			else
				tinsert(info,filterName)
			end
		end
	end
	if next(info) then
		return table.concat(info,", ")
	end
end

--[[ journal filter manipulation ]]

-- since there is no C_PetJournal.GetSearchFilter, we need to watch for it changing
hooksecurefunc(C_PetJournal,"SetSearchFilter",function(search)
	if not roster.updatingSelf then
		roster.journalBackup.search = search or ""
	end
end)
hooksecurefunc(C_PetJournal,"ClearSearchFilter",function(...)
	if not roster.updatingSelf then
		roster.journalBackup.search = ""
	end
end)

-- backs up filters and expands journal to show all pets
function roster:ExpandJournalFilters()
	roster.updatingSelf = true
	-- we're going to be triggering a PJLU event, expand disables event; restore turns it on after a delay
	rematch:UnregisterEvent("PET_JOURNAL_LIST_UPDATE")
	-- backup all filters
	local backup = roster.journalBackup
	backup.collected = not C_PetJournal.IsFlagFiltered(LE_PET_JOURNAL_FLAG_COLLECTED)
	backup.notCollected = not C_PetJournal.IsFlagFiltered(LE_PET_JOURNAL_FLAG_NOT_COLLECTED)
	for i=1,C_PetJournal.GetNumPetTypes() do
		backup.types[i] = not C_PetJournal.IsPetTypeFiltered(i)
	end
	for i=1,C_PetJournal.GetNumPetSources() do
		backup.sources[i] = not C_PetJournal.IsPetSourceFiltered(i)
	end
	-- expand all filters
	C_PetJournal.ClearSearchFilter()
	C_PetJournal.SetFlagFilter(LE_PET_JOURNAL_FLAG_COLLECTED,true)
	C_PetJournal.SetFlagFilter(LE_PET_JOURNAL_FLAG_NOT_COLLECTED,true)
	C_PetJournal.AddAllPetSourcesFilter()
	C_PetJournal.AddAllPetTypesFilter()
end

-- restores journal filters to the state in roster.journalBackup
function roster:RestoreJournalFilters()
	local backup = roster.journalBackup
	C_PetJournal.SetFlagFilter(LE_PET_JOURNAL_FLAG_COLLECTED,backup.collected)
	C_PetJournal.SetFlagFilter(LE_PET_JOURNAL_FLAG_NOT_COLLECTED,backup.notCollected)
	for i=1,C_PetJournal.GetNumPetSources() do
		C_PetJournal.SetPetSourceFilter(i,backup.sources[i])
	end
	for i=1,C_PetJournal.GetNumPetTypes() do
		C_PetJournal.SetPetTypeFilter(i,backup.sources[i])
	end
	C_PetJournal.SetSearchFilter(backup.search)
	roster.updatingSelf = nil
	-- we're doing messing with PJLU-triggering stuff, start event again in a second when things settle
	C_Timer.After(1,function() rematch:RegisterEvent("PET_JOURNAL_LIST_UPDATE") end)
end

--[[ on-demand stuff ]]

-- returns the source (as a number 1-10; Drop, Quest, Vendor, etc) of a speciesID.
-- unfortunately the only way to get this number is by setting the journal filters and
-- looking at what's left. this caches first encounter so later use doesn't touch filters.
function roster:GetSpeciesSource(speciesID)
	local sources = roster.speciesSources
	if not next(sources) then -- haven't collected sources this session
		roster:ExpandJournalFilters()
		for source=1,C_PetJournal.GetNumPetSources() do -- going through all pet sources
			-- set source filter to single category
			for i=1,10 do
			  C_PetJournal.SetPetSourceFilter(i,i==source)
			end
			-- fill speciesSources with the results
			for i=1,C_PetJournal.GetNumPets() do
			  local _,speciesID = C_PetJournal.GetPetInfoByIndex(i)
				if not sources[speciesID] then
					sources[speciesID] = source
				end
			end
		end
		roster:RestoreJournalFilters() -- put everything back how it was
	end
	-- after the first time sources are gathered, we just return the cached source index
	return sources[speciesID]
end

-- ensures abilities are only pulled once in a search (can be from strong vs, moveset and search)
function roster:PullAbilities(speciesID)
	if not next(rematch.abilityList) then
		C_PetJournal.GetPetAbilityList(speciesID,rematch.abilityList,rematch.levelList)
	end
end

-- returns true if the given speciesID is strong to the filtered pet types
function roster:IsStrong(speciesID)
	-- if we've done this work before, return prior result
	if roster.cacheStrong[speciesID]~=nil then
		return roster.cacheStrong[speciesID]
	end
	roster:PullAbilities(speciesID)
	local info = rematch.info
	wipe(info)
	-- fill info lookup table with abilityTypes of attacks
	for _,abilityID in ipairs(rematch.abilityList) do
		local abilityType, noHints = select(7,C_PetBattles.GetAbilityInfoByID(abilityID))
		if not noHints then
			info[abilityType] = true
		end
	end
	-- go through each filter and make sure each are accounted for in the pet's abilities
	for attackType in pairs(settings.Filters.Strong) do
		if not info[rematch.hintsDefense[attackType][1]] then
			roster.cacheStrong[speciesID] = false
			return false
		end
	end
	roster.cacheStrong[speciesID] = true
	return true
end

-- fills roster.movesetsAt25 if MovesetNot25 filter used
function roster:PopulateMovesetsAt25()
	if roster:GetFilter("Level","MovesetNot25") then
		local c = 0
		for speciesID in pairs(roster.speciesAt25) do
			C_PetJournal.GetPetAbilityList(speciesID,rematch.abilityList,rematch.levelList)
			local moveset = table.concat(rematch.abilityList,",")
			roster.movesetsAt25[moveset] = true
			c=c+1
		end
	end
end

-- returns true/false whether moveset of speciesID is at 25; uses similar cache as strong vs
function roster:IsMovesetAt25(speciesID)
	if roster.cacheMoveset[speciesID]~=nil then
		return roster.cacheMoveset[speciesID]
	end
	roster:PullAbilities(speciesID)
	local moveset = table.concat(rematch.abilityList,",")
	if roster.movesetsAt25[moveset] then
		roster.cacheMoveset[speciesID] = true
		return true
	else
		roster.cacheMoveset[speciesID] = false
		return false
	end
end

-- returns true if petID (which can be a speciesID too) is in a team
function roster:IsPetInTeam(petID)
	local sanctuary = settings.Sanctuary[petID]
	return sanctuary and sanctuary[1] and sanctuary[1]>0
end

--[[ roster update ]]

-- this should be the standard way of updating the roster; wait a frame before doing actual update
function rematch:UpdateRoster()
	if select(2,C_PetJournal.GetNumPets())==0 then -- if no pets owned, come back in half a second and try again
		rematch:StartTimer("RosterUpdate",0.5,rematch.UpdateRoster)
	else
		rematch:StartTimer("RosterUpdate",0,roster.UpdateNow)
	end
end

-- should only be called by above on a one-frame delay, or in a PetPanel udpate when number of owned pets changes
function roster:UpdateNow()
	if not roster:CheckForUpdate() then
		roster:PopulateList()
	end
	rematch:UpdateUI()
end

-- updates owned pets when necessary; should be called before dealing with roster outside pet panel
function roster:CheckForUpdate()
	if roster.needsUpdated then
		roster:UpdateOwned()
		roster.needsUpdated = nil
		roster:PopulateList()

		-- if we're here and no owned pets, then user summoned a rematch window before pets
		-- are loaded; on /reload a PET_JOURNAL_LIST_UPDATE doesn't always fire after pets
		-- load. This will check back once a second for up to five seconds.
		if select(2,C_PetJournal.GetNumPets())==0 and (roster.ownedTimeout or 0)<5 then
			C_Timer.After(1,function()
				roster.ownedTimeout = (roster.ownedTimeout or 0)+1
				roster.needsUpdated = true
				roster:CheckForUpdate()
			end)
		end

		return true
	end
end

-- this populates roster.pets with all owned and missing pets
-- it only happens when the number of pets a player owns changes
-- the sort of this list is not important until a substitute pet is sought
-- while here, also count the number of unique pets
function roster:UpdateOwned()
	wipe(roster.pets)
	roster:ExpandJournalFilters()
	wipe(rematch.info)
	wipe(roster.speciesAt25)
	-- gather all pets into roster.pets
	for i=1,C_PetJournal.GetNumPets() do
		local petID,speciesID,_,_,level = C_PetJournal.GetPetInfoByIndex(i)
		tinsert(roster.pets,petID or speciesID)
		if petID then -- if this is an owned pet, note its species
			rematch.info[speciesID] = true
			if level==25 then
				roster.speciesAt25[speciesID] = true
			end
		end
	end
	roster.uniquePets = 0
	for speciesID in pairs(rematch.info) do
		roster.uniquePets = roster.uniquePets + 1
	end
	wipe(rematch.info)
	roster:RestoreJournalFilters()
end

-- fills and sorts roster.list with pets depending on our filter settings
-- should be called in reaction to PET_JOURNAL_LIST_UPDATE
function roster:PopulateList()
	wipe(roster.list)

	roster:PopulateMovesetsAt25() -- gathers movesets only if MovesetNot25 used

	for index,petID in ipairs(roster.pets) do
		local include,sortStat,isFavorite,name = roster:FilterPetByIndex(index)
		if include then
			roster.favorites[petID] = isFavorite
			roster.sortStats[petID] = sortStat
			tinsert(roster.list,petID)
		end
	end

	-- sort pets
	local order = settings.Sort.Order
	local reversed = settings.Sort.Reverse
	if (order==1 or order==4) then -- sort by name and type are ascending
		roster.sortLowToHigh = not reversed
	else -- sort by level, rarity, health, power and speed is descending
		roster.sortLowToHigh = reversed
	end
	table.sort(roster.list,roster.SortPets)

	-- cleanup of temporary tables (many of these are empty most of the time)
	wipe(roster.movesetsAt25)
	wipe(roster.cacheStrong)
	wipe(roster.cacheMoveset)
	wipe(roster.cacheSearchSpecies)
	wipe(roster.cacheSearchAbilities)
	wipe(roster.sortStats)
	wipe(roster.favorites)
end

-- the real work done here: returns true if the index into roster.pets is to be listed
-- the quickest checks are done first and false returned at earliest opportunity
function roster:FilterPetByIndex(index)
	local petID = roster.pets[index]

	wipe(rematch.abilityList) -- PullAbilityList checks if this is empty before gathering abilities

	local speciesID, customName, level, isFavorite, name, petType, sourceText, canBattle, tradable
	local _, maxHealth, power, speed, rarity
	if type(petID)=="string" then -- this is an owned petID
		speciesID, customName, level, _, _, _, isFavorite, name, _, petType, _, sourceText, _, _, canBattle, tradable = C_PetJournal.GetPetInfoByPetID(petID)
		_, maxHealth, power, speed, rarity = C_PetJournal.GetPetStats(petID)
		if level==25 then roster.speciesAt25[speciesID] = true end -- while we're here
	else -- this is a speciesID
		name, _, petType, _, sourceText, _, _, canBattle, tradable = C_PetJournal.GetPetInfoBySpeciesID(petID)
		speciesID = petID
		petID = nil
	end

	if roster:IsFilterUsed("Collected") then
		if petID and roster:GetFilter("Collected","Owned") then return false end
		if not petID and roster:GetFilter("Collected","Missing") then return false end
	end

	if roster:IsFilterUsed("Favorite") and not isFavorite then return false end

	if roster:IsFilterUsed("Types") then
		if not roster:GetFilter("Types",petType) then return false end
	end

	if roster:IsFilterUsed("Rarity") then
		if not roster:GetFilter("Rarity",rarity) then return false end
	end

	if roster:IsFilterUsed("Sources") then
		local source = roster:GetSpeciesSource(speciesID)
		if not roster:GetFilter("Sources",source) then return false end
	end

	if roster:IsFilterUsed("Tough") then
		if not roster:GetFilter("Tough",rematch.hintsDefense[petType][2]) then return false end
	end

	if roster:IsFilterUsed("Level") then
		local pass
		if roster:GetFilter("Level","Without25s") and canBattle then if roster.speciesAt25[speciesID] then return false end pass=true
		elseif roster:GetFilter("Level","MovesetNot25") and canBattle then if roster:IsMovesetAt25(speciesID) then return false end pass=true
		elseif not petID or not canBattle then return false
		elseif roster:GetFilter("Level",4) and level==25 then pass=true
		elseif roster:GetFilter("Level",1) and level>0 and level<8 then pass=true
		elseif roster:GetFilter("Level",2) and level>7 and level<15 then pass=true
		elseif roster:GetFilter("Level",3) and level>14 and level<25 then pass=true
		end
		if not pass then return false end
	end

	if roster:IsFilterUsed("Other") then
		if roster:GetFilter("Other","Tradable") and not tradable then return false
		elseif roster:GetFilter("Other","NotTradable") and tradable then return false end
		if roster:GetFilter("Other","Battle") and not canBattle then return false
		elseif roster:GetFilter("Other","NotBattle") and canBattle then return false end

		local isLeveling = rematch:IsPetLeveling(petID)
		if roster:GetFilter("Other","Leveling") and not isLeveling then return false
		elseif roster:GetFilter("Other","NotLeveling") and isLeveling then return false end

		if roster:GetFilter("Other","InTeam") and not roster:IsPetInTeam(petID or speciesID) then return false
		elseif roster:GetFilter("Other","NotInTeam") and roster:IsPetInTeam(petID or speciesID) then return false end

		local qty = roster:GetFilter("Other","Qty1") and 1 or roster:GetFilter("Other","Qty2") and 2 or roster:GetFilter("Other","Qty3") and 3
		if qty then
			local count = C_PetJournal.GetNumCollectedInfo(speciesID)
			if not count then return false
			elseif qty==3 and count<3 then return false
			elseif qty==2 and count<2 then return false
			elseif qty==1 and count~=1 then return false
			end
		end
	end

	if roster:IsFilterUsed("Breed") then
		if not petID then return false end -- not an owned pet, it has no breed
		-- these compare the addon-supplied breed to the filter index
		if rematch.breedSource=="BattlePetBreedID" then
			-- BPBID has a lookup table; an unknown breed is 11th index into lookup table
			if not roster:GetFilter("Breed",rematch.breedLookup[GetBreedID_Journal(petID)] or 11) then return false end
		elseif rematch.breedSource=="PetTracker_Breeds" then
			-- PetTracker returns a number 3-12 for breed; subtract 2 to get filter index (unknown starts at 13-2=11)
			if not roster:GetFilter("Breed",(PetTracker.Journal:GetBreed(petID) or 13)-2) then return false end
		end
	end

	if roster:IsFilterUsed("Strong") then
		if not roster:IsStrong(speciesID) then return false end
	end

	if roster.searchMask and not roster:IsSearchMatch(speciesID,customName,name,sourceText,settings.PetNotes[speciesID]) then
		return false
	end

	if next(roster.searchStatRanges) then
		if not petID or not canBattle then return end -- only owned pets have valid stat ranges
		for stat,values in pairs(roster.searchStatRanges) do
			local liveStat = stat=="Level" and level or stat=="Health" and maxHealth or stat=="Power" and power or stat=="Speed" and speed
			if liveStat<values[1] or liveStat>values[2] then
				return false
			end
		end
	end

	if roster:IsFilterUsed("Similar") then
		roster:PullAbilities(speciesID)
		local matches = 0
		for _,abilityID in pairs(rematch.abilityList) do
			if roster:GetFilter("Similar",abilityID) then
				matches = matches + 1
			end
			if matches>2 then break end
		end
		if matches<3 then
			return false
		end
	end

	-- if we reached here, pet is to be listed, get its sort stat
	local sortOrder = settings.Sort.Order
	local sortStat
	if sortOrder==1 then sortStat = name
	elseif sortOrder==2 then sortStat = level
	elseif sortOrder==3 then sortStat = rarity
	elseif sortOrder==4 then sortStat = petType
	elseif sortOrder==5 then sortStat = maxHealth
	elseif sortOrder==6 then sortStat = power
	elseif sortOrder==7 then sortStat = speed
	end

	-- if we made it here, pet is to be listed!
	return true,sortStat,isFavorite and settings.Sort.FavoritesFirst or nil
end

--[[ search ]]

-- sets search to the search text and makes a case-insensitive/magic-char-escaped search mask
function roster:SetSearch(text)
	wipe(roster.searchStatRanges)
	if not text or text:len()==0 then
		roster.searchMask = nil
	else
		text = roster:ParseSearchStatRanges(text)
		roster.searchMask = rematch:DesensitizeText(text)
	end
end

-- returns true if the passed parameters (customName,name,sourceText) or species abilities
-- contain a match to roster.searchMask; uses two caches to drastically speed up processing
-- this is the only search function here that runs during the search
function roster:IsSearchMatch(speciesID,...)
	local mask = roster.searchMask
	for i=1,select("#",...) do
		local candidate = select(i,...)
		if candidate and candidate:match(mask) then
			return true
		end
	end
	-- searching abilities now; using caches to speed things up
	if roster.cacheSearchSpecies[speciesID]~=nil then
		return roster.cacheSearchSpecies[speciesID]
	end
	roster:PullAbilities(speciesID)
	for _,abilityID in ipairs(rematch.abilityList) do
		if roster.cacheSearchAbilities[abilityID]~=nil then
			if roster.cacheSearchAbilities[abilityID] then
				roster.cacheSearchSpecies[speciesID] = true
				return true
			end
		else
			local _,name,_,_,description = C_PetBattles.GetAbilityInfoByID(abilityID)
			if name:match(mask) or description:match(mask) then
				roster.cacheSearchAbilities[abilityID] = true
				roster.cacheSearchSpecies[speciesID] = true
				return true
			end
			roster.cacheSearchAbilities[abilityID] = false
		end
	end
	roster.cacheSearchSpecies[speciesID] = false
	return false
end

-- takes raw text from the search box and pulls out operations on stats and returns remainder
function roster:ParseSearchStatRanges(text)
	local ranges = roster.searchStatRanges
	local info = rematch.info
	wipe(info)
	local hasStatRange
	-- if any operations found (level<25, 1-24, health=100-500, etc) then pull them out
	if text:match("[<>=]%d") or text:match("%d%-%d") then
		-- fill info with clumps of text without spaces
		for operation in text:gmatch("[^ ]+") do
			tinsert(info,operation)
		end
		-- go through each clump of text and test if it's an operation
		for i=#info,1,-1 do
			local operation = info[i]
			if operation:match("[<>=%-]") then -- if this has one of < > = or - then
				stat,minValue,maxValue = roster:ParseOperation(operation:lower())
				if stat then -- operation was a stat range, add it to ranges
					ranges[stat] = {minValue,maxValue}
					hasStatRange = true
					tremove(info,i) -- and remove operation from the table of clumps
				end
			end
		end
	end
	if hasStatRange then -- some stat ranges were pulled out, rebuild text from remaining clumps
		return table.concat(info," ")
	else -- no stat ranges were pulled out, return text as it was given
		return text
	end
end

-- tests potentional operation for any stat equations and returns stat and min/max value if so
function roster:ParseOperation(operation)
	-- legacy operation: <25 or >1 or =25 for level ranges
	local operator,value = operation:match("^([<>=])(%d+)$")
	if operator then
		value = tonumber(value)
		if operator=="=" then return "Level",value,value
		elseif operator=="<" then return "Level",1,value-1
		elseif operator==">" then return "Level",value+1,25
		end
	end
	-- legacy operation: 10-11, 1-24 for level ranges
	local minValue,maxValue = operation:match("^(%d+)%-(%d+)$")
	if minValue then
		return "Level",tonumber(minValue),tonumber(maxValue)
	end
	-- stat<value or stat>value or stat=value: level<25, speed>280, power=100
	local stat,operator,value = operation:match("^(%w+)([<>=])(%d+)$")
	if stat and roster.searchStatMasks[stat] then
		stat = roster.searchStatMasks[stat]
		value = tonumber(value)
		if operator=="=" then return stat,value,value
		elseif operator=="<" then return stat,1,value-1
		elseif operator==">" then return stat,value+1,9999
		end
	end
	-- stat=min-max: level=1-24, health=200-500
	local stat,minValue,maxValue = operation:match("^(%w+)=(%d+)%-(%d+)$")
	if stat and roster.searchStatMasks[stat] then
		stat = roster.searchStatMasks[stat]
		return stat,tonumber(minValue),tonumber(maxValue)
	end
end

-- table.sort function to sort pets so:
-- owned (petID instead of speciesID) always list first
-- then favorites list before non-favorites (if FavoritesFirst true)
-- then by sort stat filled during filter
-- then by name pulled on demand (compromise between memory vs speed)
-- and lastly by the raw petID so order remains fixed
function roster.SortPets(pet1,pet2)

	-- first sort by the existence of a pet
	if pet1 and not pet2 then return true
	elseif pet2 and not pet1 then return false
	elseif not pet1 and not pet2 then return false
	end

	local o1,o2 = type(pet1)=="string", type(pet2)=="string" -- owned pets always list first
	if o1 and not o2 then return true
	elseif o2 and not o1 then return false
	end

	local f1,f2 = roster.favorites[pet1], roster.favorites[pet2] -- favorites
	if f1 and not f2 then return true -- f1/f2 only true if FavoritesFirst true
	elseif f2 and not f1 then return false
	end

	local s1,s2 = roster.sortStats[pet1], roster.sortStats[pet2] -- chosen sort stat existence
	if s1 and not s2 then return true
	elseif s2 and not s1 then return false
	elseif not s1 and not s2 then return false
	end

	-- both pets have a sort stat here; sort them depending on sortLowToHigh
	if roster.sortLowToHigh then
		if s1<s2 then return true
		elseif s2<s1 then return false
		end
	else -- level, rarity, health, power and speed sort in descending value (25-1)
		if s1>s2 then return true
		elseif s2>s1 then return false
		end
	end

	-- if we're still here, the two pets have the same sort stat

	-- gather name and level of the two pets
	local n1,n2,l1,l2,_
	if o1 then -- these are owned pets
		_,_,l1,_,_,_,_,n1 = C_PetJournal.GetPetInfoByPetID(pet1)
		_,_,l2,_,_,_,_,n2 = C_PetJournal.GetPetInfoByPetID(pet2)
	else -- unowned pets (0 level)
		l1,n1 = 0,C_PetJournal.GetPetInfoBySpeciesID(pet1)
		l2,n2 = 0,C_PetJournal.GetPetInfoBySpeciesID(pet2)
	end

	-- if sorting by rarity, then do a secondary sort by level
	if settings.Sort.Order==3 then
		if l1~=l2 then
			return l1>l2 -- always descending level
		end
	end

	-- sort stats are identical, sort by name first
--	local n1 = o1 and select(8,C_PetJournal.GetPetInfoByPetID(pet1)) or C_PetJournal.GetPetInfoBySpeciesID(pet1)
--	local n2 = o2 and select(8,C_PetJournal.GetPetInfoByPetID(pet2)) or C_PetJournal.GetPetInfoBySpeciesID(pet2)

	-- final sort: by name (or by petID if name is identical)
	if n1==n2 then
		return pet1<pet2 -- names identical; sort by petID so sort order remains fixed
	else
		return n1<n2 -- names are different; sort by name
	end
end

--[[ Similar ]]

function roster:SetSimilarFilter(speciesID)
	roster:ClearAllFilters() -- start search for similar with other filters clear
	wipe(rematch.abilityList)
	C_PetJournal.GetPetAbilityList(speciesID, rematch.abilityList, rematch.levelList)
	for _,abilityID in ipairs(rematch.abilityList) do
		roster:SetFilter("Similar",abilityID,true)
	end
	rematch:ShowPets()
	rematch:UpdateRoster()
end

